package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.config.annotations.RequireDatabase
import dev.fyloz.colorrecipesexplorer.dtos.MaterialDto
import dev.fyloz.colorrecipesexplorer.dtos.MaterialQuantityDto
import dev.fyloz.colorrecipesexplorer.dtos.MixDeductDto
import dev.fyloz.colorrecipesexplorer.dtos.MixMaterialDto
import dev.fyloz.colorrecipesexplorer.exception.RestException
import dev.fyloz.colorrecipesexplorer.model.Material
import dev.fyloz.colorrecipesexplorer.utils.mapMayThrow
import org.springframework.http.HttpStatus
import org.springframework.stereotype.Service
import javax.transaction.Transactional

interface InventoryLogic {
    /** Adds each given [MaterialQuantityDto] to the inventory and returns the updated quantities. */
    fun add(materialQuantities: Collection<MaterialQuantityDto>): Collection<MaterialQuantityDto>

    /** Adds a given quantity to the given [Material]'s inventory quantity according to the given [materialQuantity] and returns the updated quantity. */
    fun add(materialQuantity: MaterialQuantityDto): Float

    /** Deducts the inventory quantity of each [Material]s in the mix according to the ratio defined in the given [mixRatio] and returns the updated quantities. */
    fun deductMix(mixRatio: MixDeductDto): Collection<MaterialQuantityDto>

    /** Deducts the inventory quantity of each given [MaterialQuantityDto] and returns the updated quantities. */
    fun deduct(materialQuantities: Collection<MaterialQuantityDto>): Collection<MaterialQuantityDto>

    /** Deducts the inventory quantity of a given [Material] by a given quantity according to the given [materialQuantity] and returns the updated quantity. */
    fun deduct(materialQuantity: MaterialQuantityDto): Float
}

@Service
@RequireDatabase
class DefaultInventoryLogic(
    private val materialLogic: MaterialLogic,
    private val mixLogic: MixLogic
) : InventoryLogic {
    @Transactional
    override fun add(materialQuantities: Collection<MaterialQuantityDto>) =
        materialQuantities.map { MaterialQuantityDto(it.materialId, add(it)) }

    override fun add(materialQuantity: MaterialQuantityDto) =
        materialLogic.updateQuantity(
            materialLogic.getById(materialQuantity.materialId),
            materialQuantity.quantity
        )

    @Transactional
    override fun deductMix(mixRatio: MixDeductDto): Collection<MaterialQuantityDto> {
        val mix = mixLogic.getById(mixRatio.id)
        val mixMaterials = mix.mixQuantities.materials

        if (mixMaterials.isEmpty()) return listOf()
        return deduct(getMaterialsWithAdjustedQuantities(mixMaterials, mixRatio))
    }

    @Transactional
    override fun deduct(materialQuantities: Collection<MaterialQuantityDto>): Collection<MaterialQuantityDto> {
        val thrown = mutableListOf<NotEnoughInventoryException>()

        val updatedQuantities =
            materialQuantities.mapMayThrow<MaterialQuantityDto, MaterialQuantityDto, NotEnoughInventoryException>(
                { thrown.add(it) }
            ) {
                MaterialQuantityDto(it.materialId, deduct(it))
            }

        if (thrown.isNotEmpty()) {
            throw MultiplesNotEnoughInventoryException(thrown)
        }

        return updatedQuantities
    }

    override fun deduct(materialQuantity: MaterialQuantityDto): Float =
        with(materialLogic.getById(materialQuantity.materialId)) {
            if (this.inventoryQuantity >= materialQuantity.quantity) {
                materialLogic.updateQuantity(this, -materialQuantity.quantity)
            } else {
                throw NotEnoughInventoryException(materialQuantity.quantity, this)
            }
        }

    private fun getMaterialsWithAdjustedQuantities(
        mixMaterials: Collection<MixMaterialDto>,
        mixRatio: MixDeductDto
    ): Collection<MaterialQuantityDto> {
        val adjustedFirstMaterialQuantity = mixMaterials.first().quantity * mixRatio.ratio

        fun getAdjustedQuantity(material: MaterialDto, quantity: Float) =
            if (!material.materialType.usePercentages)
                quantity * mixRatio.ratio   // Simply multiply the quantity by the ratio
            else
                (quantity * adjustedFirstMaterialQuantity) / 100f   // Percents quantities are a ratio of the first material

        return mixMaterials.associate { it.material to it.quantity }
            .mapValues { getAdjustedQuantity(it.key, it.value) }
            .map { MaterialQuantityDto(it.key.id, it.value) }
    }
}

class NotEnoughInventoryException(quantity: Float, material: MaterialDto) :
    RestException(
        "notenoughinventory",
        "Not enough inventory",
        HttpStatus.BAD_REQUEST,
        "Cannot deduct ${quantity}mL of ${material.name} because there is only ${material.inventoryQuantity}mL in inventory",
        mapOf(
            "material" to material.name,
            "materialId" to material.id.toString(),
            "requestQuantity" to quantity,
            "availableQuantity" to material.inventoryQuantity
        )
    )

class MultiplesNotEnoughInventoryException(exceptions: List<NotEnoughInventoryException>) :
    RestException(
        "notenoughinventory-multiple",
        "Not enough inventory",
        HttpStatus.BAD_REQUEST,
        "Cannot deduct requested quantities because there is no enough of them in inventory",
        mapOf(
            "lowQuantities" to exceptions.map { it.extensions }
        )
    )
