package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.config.annotations.LogicComponent
import dev.fyloz.colorrecipesexplorer.dtos.MaterialDto
import dev.fyloz.colorrecipesexplorer.dtos.MaterialSaveDto
import dev.fyloz.colorrecipesexplorer.dtos.MixTypeDto
import dev.fyloz.colorrecipesexplorer.logic.files.WriteableFileLogic
import dev.fyloz.colorrecipesexplorer.model.Material
import dev.fyloz.colorrecipesexplorer.service.MaterialService

interface MaterialLogic : Logic<MaterialDto, MaterialService> {
    /** Checks if a material with the given [name] exists. */
    fun existsByName(name: String): Boolean

    /**
     * Returns every material available in the context of the recipe with the given [recipeId].
     * The materials included contains every non mix type material, and the materials generated for the recipe mix types.
     */
    fun getAllForRecipe(recipeId: Long): Collection<MaterialDto>

    /**
     * Returns every material available in the context of the mix with the given [mixId].
     * The materials included contains every non mix type material, and the materials generated for
     * the mix's recipe mix types, excluding the mix's mix type.
     */
    fun getAllForMix(mixId: Long): Collection<MaterialDto>

    /** Saves the given [dto]. */
    fun save(dto: MaterialSaveDto): MaterialDto

    /** Updates the given [dto]. */
    fun update(dto: MaterialSaveDto): MaterialDto

    /** Updates the quantity of the given [material] with the given [factor] and returns the updated quantity. */
    fun updateQuantity(material: MaterialDto, factor: Float): Float
}

@LogicComponent
class DefaultMaterialLogic(
    service: MaterialService,
    val recipeLogic: RecipeLogic,
    val mixLogic: MixLogic,
    val materialTypeLogic: MaterialTypeLogic,
    val fileLogic: WriteableFileLogic
) : BaseLogic<MaterialDto, MaterialService>(service, Constants.ModelNames.MATERIAL), MaterialLogic {
    override fun existsByName(name: String) = service.existsByName(name, null)

    override fun getAllForRecipe(recipeId: Long): Collection<MaterialDto> {
        val recipe = recipeLogic.getById(recipeId)

        return getAllWithMixTypesMaterials(recipe.mixTypes)
    }

    override fun getAllForMix(mixId: Long): Collection<MaterialDto> {
        val mix = mixLogic.getById(mixId)
        val recipe = recipeLogic.getById(mix.recipeId)

        val availableMixTypes = recipe.mixTypes.filter { it != mix.mixType }
        return getAllWithMixTypesMaterials(availableMixTypes)
    }

    private fun getAllWithMixTypesMaterials(mixTypes: Collection<MixTypeDto>) =
        getAll() + mixTypes.map { it.asMaterial() }

    override fun save(dto: MaterialSaveDto) = save(saveDtoToDto(dto, false)).also { saveSimdutFile(dto, false) }
    override fun save(dto: MaterialDto): MaterialDto {
        throwIfNameAlreadyExists(dto.name)

        return super.save(dto)
    }

    override fun update(dto: MaterialSaveDto) = update(saveDtoToDto(dto, true)).also { saveSimdutFile(dto, true) }
    override fun update(dto: MaterialDto): MaterialDto {
        throwIfNameAlreadyExists(dto.name, dto.id)

        return super.update(dto)
    }

    override fun updateQuantity(material: MaterialDto, factor: Float): Float {
        val updatedQuantity = material.inventoryQuantity + factor
        service.updateInventoryQuantityById(material.id, updatedQuantity)

        return updatedQuantity
    }

    override fun deleteById(id: Long) {
        if (service.isUsedByMixMaterialOrMixType(id)) {
            throw cannotDeleteException("Cannot delete the material with the id '$id' because mix types and/or recipes depends on it")
        }

        val material = getById(id)
        val simdutPath = Material.getSimdutFilePath(material.name)
        if (fileLogic.exists(simdutPath)) {
            fileLogic.delete(simdutPath)
        }

        super.deleteById(id)
    }

    private fun saveDtoToDto(saveDto: MaterialSaveDto, updating: Boolean): MaterialDto {
        val isMixType = !updating || getById(saveDto.id).isMixType
        val materialType = materialTypeLogic.getById(saveDto.materialTypeId)

        return MaterialDto(saveDto.id, saveDto.name, saveDto.inventoryQuantity, isMixType, materialType)
    }

    private fun saveSimdutFile(dto: MaterialSaveDto, updating: Boolean) {
        val file = dto.simdutFile

        if (file != null && !file.isEmpty) {
            fileLogic.write(file, Material.getSimdutFilePath(dto.name), updating)
        }
    }

    private fun throwIfNameAlreadyExists(name: String, id: Long? = null) {
        if (service.existsByName(name, id)) {
            throw alreadyExistsException(value = name)
        }
    }
}