package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.config.annotations.LogicComponent
import dev.fyloz.colorrecipesexplorer.dtos.*
import dev.fyloz.colorrecipesexplorer.exception.InvalidPositionError
import dev.fyloz.colorrecipesexplorer.exception.InvalidPositionsException
import dev.fyloz.colorrecipesexplorer.exception.RestException
import dev.fyloz.colorrecipesexplorer.utils.PositionUtils
import org.springframework.context.annotation.Lazy
import org.springframework.http.HttpStatus

interface MixQuantityLogic {
    /**
     *  Validates if the given [mixMaterials]. To be valid, the position of each mix material must be greater or equals to 1 and unique in the set.
     *  There must also be no gap between the positions. Also, the quantity of the first mix material in the set must not be expressed in percentages.
     *  If any of those criteria are not met, an [InvalidGroupStepsPositionsException] will be thrown.
     */
    fun validateMixQuantities(mixMaterials: List<MixQuantityDto>)

    /** Validates the given mix quantities [dtos] and put them in [MixQuantitiesDto] to be consumed by a mix. */
    fun validateAndPrepareForMix(dtos: List<MixQuantitySaveDto>): MixQuantitiesDto
}

@LogicComponent
class DefaultMixQuantityLogic(
    @Lazy private val materialLogic: MaterialLogic,
    private val mixTypeLogic: MixTypeLogic
) : MixQuantityLogic {
    override fun validateMixQuantities(mixMaterials: List<MixQuantityDto>) {
        if (mixMaterials.isEmpty()) return

        val sortedMixMaterials = mixMaterials.sortedBy { it.position }

        try {
            PositionUtils.validate(sortedMixMaterials.map { it.position })
        } catch (ex: InvalidPositionsException) {
            throw InvalidMixMaterialsPositionsException(ex.errors)
        }

        val firstMixMaterial = sortedMixMaterials[0]
        if (firstMixMaterial is MixMaterialDto) {
            if (firstMixMaterial.material.materialType.usePercentages) {
                throw InvalidFirstMixMaterialException(sortedMixMaterials[0])
            }
        }
    }

    override fun validateAndPrepareForMix(dtos: List<MixQuantitySaveDto>): MixQuantitiesDto {
        val mixMixTypes = dtos.filter { it.isMixType }.map {
            MixMixTypeDto(
                id = it.id,
                mixType = mixTypeLogic.getById(it.materialId),
                quantity = it.quantity,
                position = it.position
            )
        }

        val mixMaterials = dtos.filter { !it.isMixType }.map {
            MixMaterialDto(
                id = it.id,
                material = materialLogic.getById(it.materialId),
                quantity = it.quantity,
                position = it.position
            )
        }

        validateMixQuantities(mixMixTypes + mixMaterials)
        return MixQuantitiesDto(mixMaterials, mixMixTypes)
    }
}

// TODO check if required
class InvalidMixMaterialsPositionsException(
    val errors: Set<InvalidPositionError>
) : RestException(
    "invalid-mixmaterial-position",
    "Invalid mix materials positions",
    HttpStatus.BAD_REQUEST,
    "The position of mix materials are invalid",
    mapOf(
        "invalidMixMaterials" to errors
    )
)

class InvalidFirstMixMaterialException(
    val mixMaterial: MixQuantityDto
) : RestException(
    "invalid-mixmaterial-first",
    "Invalid first mix material",
    HttpStatus.BAD_REQUEST,
    "The first mix material is invalid because its material must not be expressed in percents",
    mapOf(
        "mixMaterial" to mixMaterial
    )
)