package dev.fyloz.colorrecipesexplorer.rest

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeEditUsers
import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeViewUsers
import dev.fyloz.colorrecipesexplorer.dtos.GroupDto
import dev.fyloz.colorrecipesexplorer.dtos.UserDto
import dev.fyloz.colorrecipesexplorer.dtos.UserSaveDto
import dev.fyloz.colorrecipesexplorer.dtos.UserUpdateDto
import dev.fyloz.colorrecipesexplorer.logic.users.GroupLogic
import dev.fyloz.colorrecipesexplorer.logic.users.UserLogic
import dev.fyloz.colorrecipesexplorer.model.account.Permission
import org.springframework.context.annotation.Profile
import org.springframework.http.MediaType
import org.springframework.security.access.prepost.PreAuthorize
import org.springframework.web.bind.annotation.*
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse
import javax.validation.Valid

@RestController
@RequestMapping(Constants.ControllerPaths.USER)
@Profile("!emergency")
class UserController(private val userLogic: UserLogic) {
    @GetMapping
    @PreAuthorizeViewUsers
    fun getAll() =
        ok(userLogic.getAll())

    @GetMapping("{id}")
    @PreAuthorizeViewUsers
    fun getById(@PathVariable id: Long) =
        ok(userLogic.getById(id))

    @PostMapping
    @PreAuthorizeEditUsers
    fun save(@Valid @RequestBody user: UserSaveDto) =
        created<UserDto>(Constants.ControllerPaths.USER) {
            userLogic.save(user)
        }

    @PutMapping
    @PreAuthorizeEditUsers
    fun update(@Valid @RequestBody user: UserUpdateDto) =
        noContent {
            userLogic.update(user)
        }

    @PutMapping("{id}/password", consumes = [MediaType.TEXT_PLAIN_VALUE])
    @PreAuthorizeEditUsers
    fun updatePassword(@PathVariable id: Long, @RequestBody password: String) =
        noContent {
            userLogic.updatePassword(id, password)
        }

    @PutMapping("{userId}/permissions/{permission}")
    @PreAuthorizeEditUsers
    fun addPermission(
        @PathVariable userId: Long,
        @PathVariable permission: Permission
    ) = noContent {
        userLogic.addPermission(userId, permission)
    }

    @DeleteMapping("{userId}/permissions/{permission}")
    @PreAuthorizeEditUsers
    fun removePermission(
        @PathVariable userId: Long,
        @PathVariable permission: Permission
    ) = noContent {
        userLogic.removePermission(userId, permission)
    }

    @DeleteMapping("{id}")
    @PreAuthorizeEditUsers
    fun deleteById(@PathVariable id: Long) =
        userLogic.deleteById(id)
}

@RestController
@RequestMapping(Constants.ControllerPaths.GROUP)
@Profile("!emergency")
class GroupsController(
    private val groupLogic: GroupLogic,
    private val userLogic: UserLogic
) {
    @GetMapping
    @PreAuthorize("hasAnyAuthority('VIEW_RECIPES', 'VIEW_USERS')")
    fun getAll() =
        ok(groupLogic.getAll())

    @GetMapping("{id}")
    @PreAuthorizeViewUsers
    fun getById(@PathVariable id: Long) =
        ok(groupLogic.getById(id))

    @GetMapping("{id}/users")
    @PreAuthorizeViewUsers
    fun getUsersForGroup(@PathVariable id: Long) =
        ok(groupLogic.getUsersForGroup(id))

    @PostMapping("default/{groupId}")
    @PreAuthorizeViewUsers
    fun setDefaultGroup(@PathVariable groupId: Long, response: HttpServletResponse) =
        noContent {
            groupLogic.setResponseDefaultGroup(groupId, response)
        }

    @GetMapping("default")
    @PreAuthorizeViewUsers
    fun getRequestDefaultGroup(request: HttpServletRequest) =
        ok(with(groupLogic) {
            getRequestDefaultGroup(request)
        })

    @GetMapping("currentuser")
    fun getCurrentGroupUser(request: HttpServletRequest) =
        ok(with(groupLogic.getRequestDefaultGroup(request)) {
            userLogic.getDefaultGroupUser(this)
        })

    @PostMapping
    @PreAuthorizeEditUsers
    fun save(@Valid @RequestBody group: GroupDto) =
        created<GroupDto>(Constants.ControllerPaths.GROUP) {
            groupLogic.save(group)
        }

    @PutMapping
    @PreAuthorizeEditUsers
    fun update(@Valid @RequestBody group: GroupDto) =
        noContent {
            groupLogic.update(group)
        }

    @DeleteMapping("{id}")
    @PreAuthorizeEditUsers
    fun deleteById(@PathVariable id: Long) =
        noContent {
            groupLogic.deleteById(id)
        }
}

@RestController
@RequestMapping("api")
@Profile("!emergency")
class LogoutController(private val userLogic: UserLogic) {
    @GetMapping("logout")
    @PreAuthorize("isFullyAuthenticated()")
    fun logout(request: HttpServletRequest) =
        ok {
            userLogic.logout(request)
        }
}
