package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.config.annotations.ServiceComponent
import dev.fyloz.colorrecipesexplorer.dtos.MaterialTypeDto
import dev.fyloz.colorrecipesexplorer.model.MaterialType
import dev.fyloz.colorrecipesexplorer.repository.MaterialTypeRepository

interface MaterialTypeService : Service<MaterialTypeDto, MaterialType, MaterialTypeRepository> {
    /** Checks if a system material type with the given [id] exists. */
    fun existsById(id: Long, systemType: Boolean): Boolean

    /** Checks if a material type with the given [name] and a different [id] exists. */
    fun existsByName(name: String, id: Long?): Boolean

    /** Checks if a material type with the given [prefix] and a different [id] exists. */
    fun existsByPrefix(prefix: String, id: Long?): Boolean

    /** Gets all material types which are or not a [systemType]. */
    fun getAll(systemType: Boolean): Collection<MaterialTypeDto>

    /** Gets the material type with the given [name]. */
    fun getByName(name: String): MaterialTypeDto?

    /** Checks if a material depends on the material type with the given [id]. */
    fun isUsedByMaterial(id: Long): Boolean
}

@ServiceComponent
class DefaultMaterialTypeService(repository: MaterialTypeRepository) :
    BaseService<MaterialTypeDto, MaterialType, MaterialTypeRepository>(repository), MaterialTypeService {
    override fun existsById(id: Long, systemType: Boolean) = repository.existsByIdAndSystemTypeIs(id, systemType)
    override fun existsByName(name: String, id: Long?) = repository.existsByNameAndIdNot(name, id ?: 0)
    override fun existsByPrefix(prefix: String, id: Long?) = repository.existsByPrefixAndIdNot(prefix, id ?: 0)
    override fun getAll(systemType: Boolean) = repository.findAllBySystemTypeIs(systemType).map(::toDto)
    override fun getByName(name: String) = repository.findByName(name)?.let(::toDto)

    override fun isUsedByMaterial(id: Long) = repository.isUsedByMaterial(id)

    override fun toDto(entity: MaterialType) =
        MaterialTypeDto(entity.id, entity.name, entity.prefix, entity.usePercentages, entity.systemType)

    override fun toEntity(dto: MaterialTypeDto) =
        MaterialType(dto.id, dto.name, dto.prefix, dto.usePercentages, dto.systemType)
}