package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.config.annotations.ServiceComponent
import dev.fyloz.colorrecipesexplorer.dtos.MixTypeDto
import dev.fyloz.colorrecipesexplorer.model.MixType
import dev.fyloz.colorrecipesexplorer.repository.MixTypeRepository

interface MixTypeService : Service<MixTypeDto, MixType, MixTypeRepository> {
    /** Checks if a mix type with the given [name], [materialTypeId] and a different [id] exists. */
    fun existsByNameAndMaterialType(name: String, materialTypeId: Long, id: Long? = null): Boolean

    /** Finds the mix type with the given [name] and [materialTypeId]. */
    fun getByNameAndMaterialType(name: String, materialTypeId: Long): MixTypeDto?

    /** Checks if a mix depends on the mix type with the given [id]. */
    fun isUsedByMixes(id: Long): Boolean

    /** Checks if the mix type with the given [id] is used by more than one mix. */
    fun isShared(id: Long): Boolean
}

@ServiceComponent
class DefaultMixTypeService(
    repository: MixTypeRepository,
    val materialService: MaterialService,
    val materialTypeService: MaterialTypeService
) :
    BaseService<MixTypeDto, MixType, MixTypeRepository>(repository), MixTypeService {
    override fun existsByNameAndMaterialType(name: String, materialTypeId: Long, id: Long?) =
        repository.existsByNameAndMaterialTypeAndIdNot(name, materialTypeId, id ?: 0L)

    override fun getByNameAndMaterialType(name: String, materialTypeId: Long) =
        repository.findByNameAndMaterialType(name, materialTypeId)?.let(::toDto)

    override fun isUsedByMixes(id: Long) = repository.isUsedByMixes(id)
    override fun isShared(id: Long) = repository.isShared(id)

    override fun toDto(entity: MixType) =
        MixTypeDto(
            entity.id,
            entity.name,
            materialTypeService.toDto(entity.materialType),
            if (entity.material != null) materialService.toDto(entity.material) else null
        )

    override fun toEntity(dto: MixTypeDto) =
        MixType(
            dto.id, dto.name,
            materialTypeService.toEntity(dto.materialType),
            if (dto.material != null) materialService.toEntity(dto.material) else null
        )
}