package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.config.annotations.ServiceComponent
import dev.fyloz.colorrecipesexplorer.dtos.GroupDto
import dev.fyloz.colorrecipesexplorer.dtos.UserDto
import dev.fyloz.colorrecipesexplorer.model.account.Permission
import dev.fyloz.colorrecipesexplorer.model.account.User
import dev.fyloz.colorrecipesexplorer.model.account.flat
import dev.fyloz.colorrecipesexplorer.repository.UserRepository
import org.springframework.data.repository.findByIdOrNull

interface UserService : Service<UserDto, User, UserRepository> {
    /** Checks if a user with the given [firstName] and [lastName] exists. */
    fun existsByFirstNameAndLastName(firstName: String, lastName: String, id: Long? = null): Boolean

    /** Gets all users, depending on [isSystemUser] and [isDefaultGroupUser]. */
    fun getAll(isSystemUser: Boolean, isDefaultGroupUser: Boolean): Collection<UserDto>

    /** Gets all users for the given [group]. */
    fun getAllByGroup(group: GroupDto): Collection<UserDto>

    /** Finds the user with the given [id], depending on [isSystemUser] and [isDefaultGroupUser]. */
    fun getById(id: Long, isSystemUser: Boolean, isDefaultGroupUser: Boolean): UserDto?

    /** Finds the user with the given [firstName] and [lastName]. */
    fun getByFirstNameAndLastName(firstName: String, lastName: String): UserDto?

    /** Find the default user for the given [group]. */
    fun getDefaultGroupUser(group: GroupDto): UserDto?
}

@ServiceComponent
class DefaultUserService(repository: UserRepository, private val groupService: GroupService) :
    BaseService<UserDto, User, UserRepository>(repository), UserService {
    override fun existsByFirstNameAndLastName(firstName: String, lastName: String, id: Long?) =
        repository.existsByFirstNameAndLastNameAndIdNot(firstName, lastName, id ?: 0L)

    override fun getAll(isSystemUser: Boolean, isDefaultGroupUser: Boolean) =
        repository.findAll()
            .filter { isSystemUser || !it.isSystemUser }
            .filter { isDefaultGroupUser || !it.isDefaultGroupUser }
            .map(::toDto)

    override fun getAllByGroup(group: GroupDto) =
        repository.findAllByGroup(groupService.toEntity(group))
            .map(::toDto)

    override fun getById(id: Long, isSystemUser: Boolean, isDefaultGroupUser: Boolean): UserDto? {
        val user = repository.findByIdOrNull(id) ?: return null
        if ((!isSystemUser && user.isSystemUser) ||
            !isDefaultGroupUser && user.isDefaultGroupUser
        ) {
            return null
        }

        return toDto(user)
    }

    override fun getByFirstNameAndLastName(firstName: String, lastName: String): UserDto? {
        val user = repository.findByFirstNameAndLastName(firstName, lastName)
        return if (user != null) toDto(user) else null
    }

    override fun getDefaultGroupUser(group: GroupDto): UserDto? {
        val user = repository.findDefaultGroupUser(groupService.toEntity(group))
        return if (user != null) toDto(user) else null
    }

    override fun toDto(entity: User) = UserDto(
        entity.id,
        entity.firstName,
        entity.lastName,
        entity.password,
        if (entity.group != null) groupService.toDto(entity.group) else null,
        getFlattenPermissions(entity),
        entity.permissions,
        entity.lastLoginTime,
        entity.isDefaultGroupUser,
        entity.isSystemUser
    )

    override fun toEntity(dto: UserDto) = User(
        dto.id,
        dto.firstName,
        dto.lastName,
        dto.password,
        dto.isDefaultGroupUser,
        dto.isSystemUser,
        if (dto.group != null) groupService.toEntity(dto.group) else null,
        dto.explicitPermissions,
        dto.lastLoginTime
    )

    private fun getFlattenPermissions(user: User): List<Permission> {
        val perms = user.permissions.flatMap { it.flat() }.filter { !it.deprecated }

        if (user.group != null) {
            return perms + groupService.flattenPermissions(user.group)
        }

        return perms
    }
}