package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.dtos.*
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.CannotDeleteException
import dev.fyloz.colorrecipesexplorer.logic.files.WriteableFileLogic
import dev.fyloz.colorrecipesexplorer.model.Material
import dev.fyloz.colorrecipesexplorer.service.MaterialService
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import org.springframework.mock.web.MockMultipartFile
import org.springframework.web.multipart.MultipartFile
import kotlin.test.assertContains
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

class DefaultMaterialLogicTest {
    private val materialServiceMock = mockk<MaterialService>()
    private val recipeLogicMock = mockk<RecipeLogic>()
    private val mixLogicMock = mockk<MixLogic>()
    private val materialTypeLogicMock = mockk<MaterialTypeLogic>()
    private val fileLogicMock = mockk<WriteableFileLogic>()

    private val materialLogic = spyk(
        DefaultMaterialLogic(
            materialServiceMock, recipeLogicMock, mixLogicMock, materialTypeLogicMock, fileLogicMock
        )
    )

    private val materialType = MaterialTypeDto(1L, "Unit test material type", "UNT", usePercentages = false)
    private val material = MaterialDto(1L, "Unit test material", 1000f, false, materialType)
    private val materialMixType = material.copy(id = 2L, isMixType = true)
    private val materialMixType2 = material.copy(id = 3L, isMixType = true)
    private val company = CompanyDto(1L, "Unit test company")
    private val recipe = RecipeDto(
        1L,
        "Unit test recipe",
        "Unit test recipe",
        "#FFFFFF",
        0,
        123,
        null,
        false,
        "A remark",
        company,
        mutableListOf(),
        listOf()
    )
    private val mix = MixDto(
        1L, "location", recipe.id, mixType = MixTypeDto(1L, "Unit test mix type", materialType), MixQuantitiesDto()
    )
    private val mix2 = mix.copy(id = 2L, mixType = mix.mixType.copy(id = 2L, material = materialMixType2))

    private val simdutFileMock = MockMultipartFile(
        "Unit test SIMDUT",
        byteArrayOf(1, 2, 3, 4)
    ) // Put some content in the mock file, so it is not ignored
    private val materialSaveDto = MaterialSaveDto(1L, "Unit test material", 1000f, materialType.id, simdutFileMock)

    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    @Test
    fun existsByName_normalBehavior_returnsTrue() {
        // Arrange
        every { materialServiceMock.existsByName(any(), any()) } returns true

        // Act
        val exists = materialLogic.existsByName(material.name)

        // Assert
        assertTrue(exists)
    }

    @Test
    fun existsByName_notFound_returnsFalse() {
        // Arrange
        every { materialServiceMock.existsByName(any(), any()) } returns false

        // Act
        val exists = materialLogic.existsByName(material.name)

        // Assert
        assertFalse(exists)
    }

    @Test
    fun getAllForMixCreation_normalBehavior_returnsNonMixTypeMaterials() {
        // Arrange
        every { materialLogic.getAll() } returns listOf(material, materialMixType2)
        every { recipeLogicMock.getById(any()) } returns recipe

        // Act
        val materials = materialLogic.getAllForRecipe(recipe.id)

        // Assert
        assertContains(materials, material)
    }

    @Test
    fun getAllForMixCreation_normalBehavior_returnsRecipeMixTypesMaterials() {
        // Arrange
        val recipe = recipe.copy(mixes = listOf(mix, mix2))

        every { materialLogic.getAll() } returns listOf(material, materialMixType2)
        every { recipeLogicMock.getById(any()) } returns recipe

        // Act
        val materials = materialLogic.getAllForRecipe(recipe.id)

        // Assert
        assertContains(materials, materialMixType2)
    }

    @Test
    fun getAllForMixUpdate_normalBehavior_returnsNonMixTypeMaterials() {
        // Arrange
        every { materialLogic.getAll() } returns listOf(material, materialMixType, materialMixType2)
        every { recipeLogicMock.getById(any()) } returns recipe
        every { mixLogicMock.getById(any()) } returns mix

        // Act
        val materials = materialLogic.getAllForMix(mix.id)

        // Assert
        assertContains(materials, material)
    }

    @Test
    fun getAllForMixUpdate_normalBehavior_returnsRecipeMixTypesMaterials() {
        // Arrange
        val recipe = recipe.copy(mixes = listOf(mix, mix2))
        val mix = mix.copy(recipeId = recipe.id)

        every { materialLogic.getAll() } returns listOf(material, materialMixType, materialMixType2)
        every { recipeLogicMock.getById(any()) } returns recipe
        every { mixLogicMock.getById(any()) } returns mix

        // Act
        val materials = materialLogic.getAllForMix(mix.id)

        // Assert
        assertContains(materials, materialMixType2)
    }

    @Test
    fun save_materialSaveDto_normalBehavior_callsSave() {
        // Arrange
        every { materialLogic.save(any<MaterialDto>()) } returns material
        every { materialTypeLogicMock.getById(any()) } returns materialType
        every { fileLogicMock.write(any<MultipartFile>(), any(), any()) } just runs

        // Act
        materialLogic.save(materialSaveDto)

        // Assert
        verify {
            materialLogic.save(any<MaterialDto>())
        }
    }

    @Test
    fun save_materialSaveDto_normalBehavior_callsWriteInFileService() {
        // Arrange
        every { materialLogic.save(any<MaterialDto>()) } returns material
        every { materialTypeLogicMock.getById(any()) } returns materialType
        every { fileLogicMock.write(any<MultipartFile>(), any(), any()) } just runs

        // Act
        materialLogic.save(materialSaveDto)

        // Assert
        verify {
            fileLogicMock.write(simdutFileMock, any(), false)
        }
        confirmVerified(fileLogicMock)
    }

    @Test
    fun save_materialSaveDto_noSimdutFile_doesNotCallWriteInFileService() {
        // Arrange
        every { materialLogic.save(any<MaterialDto>()) } returns material
        every { materialTypeLogicMock.getById(any()) } returns materialType
        every { fileLogicMock.write(any<MultipartFile>(), any(), any()) } just runs

        val saveDto = materialSaveDto.copy(simdutFile = null)

        // Act
        materialLogic.save(saveDto)

        // Assert
        verify(exactly = 0) {
            fileLogicMock.write(simdutFileMock, any(), false)
        }
        confirmVerified(fileLogicMock)
    }

    @Test
    fun save_nameExists_throwsNameAlreadyExists() {
        // Arrange
        every { materialServiceMock.existsByName(any(), any()) } returns true

        // Act
        // Assert
        assertThrows<AlreadyExistsException> { materialLogic.save(material) }
    }

    @Test
    fun update_saveDto_normalBehavior_callsUpdate() {
        // Arrange
        every { materialLogic.getById(any()) } returns material
        every { materialLogic.update(any<MaterialDto>()) } returns material
        every { materialTypeLogicMock.getById(any()) } returns materialType
        every { fileLogicMock.write(any<MultipartFile>(), any(), any()) } just runs

        // Act
        materialLogic.update(materialSaveDto)

        // Assert
        verify {
            materialLogic.update(any<MaterialDto>())
        }
    }

    @Test
    fun update_materialSaveDto_normalBehavior_callsWriteInFileService() {
        // Arrange
        every { materialLogic.getById(any()) } returns material
        every { materialLogic.update(any<MaterialDto>()) } returns material
        every { materialTypeLogicMock.getById(any()) } returns materialType
        every { fileLogicMock.write(any<MultipartFile>(), any(), any()) } just runs

        // Act
        materialLogic.update(materialSaveDto)

        // Assert
        verify {
            fileLogicMock.write(simdutFileMock, any(), true)
        }
        confirmVerified(fileLogicMock)
    }

    @Test
    fun update_materialSaveDto_noSimdutFile_doesNotCallWriteInFileService() {
        // Arrange
        every { materialLogic.getById(any()) } returns material
        every { materialLogic.update(any<MaterialDto>()) } returns material
        every { materialTypeLogicMock.getById(any()) } returns materialType
        every { fileLogicMock.write(any<MultipartFile>(), any(), any()) } just runs

        val saveDto = materialSaveDto.copy(simdutFile = null)

        // Act
        materialLogic.update(saveDto)

        // Assert
        verify(exactly = 0) {
            fileLogicMock.write(simdutFileMock, any(), true)
        }
        confirmVerified(fileLogicMock)
    }

    @Test
    fun updateQuantity_normalBehavior_callsUpdateInventoryQuantityByIdInService() {
        // Arrange
        every { materialServiceMock.updateInventoryQuantityById(any(), any()) } just runs

        val factor = 3f

        // Act
        materialLogic.updateQuantity(material, factor)

        // Assert
        verify {
            materialServiceMock.updateInventoryQuantityById(material.id, material.inventoryQuantity + factor)
        }
        confirmVerified(materialServiceMock)
    }

    @Test
    fun updateQuantity_normalBehavior_returnsUpdatedQuantity() {
        // Arrange
        every { materialServiceMock.updateInventoryQuantityById(any(), any()) } just runs

        val factor = 3f

        // Act
        val updatedQuantity = materialLogic.updateQuantity(material, factor)

        // Assert
        assertEquals(material.inventoryQuantity + factor, updatedQuantity)
    }

    @Test
    fun deleteById_normalBehavior_callsDeleteInFileLogic() {
        // Arrange
        every { materialLogic.getById(any()) } returns material
        every { materialServiceMock.isUsedByMixMaterialOrMixType(any()) } returns false
        every { materialServiceMock.deleteById(any()) } just runs
        every { fileLogicMock.exists(any()) } returns true
        every { fileLogicMock.delete(any()) } just runs

        val simdutPath = Material.getSimdutFilePath(material.name)

        // Act
        materialLogic.deleteById(material.id)

        // Assert
        verify {
            fileLogicMock.exists(simdutPath)
            fileLogicMock.delete(simdutPath)
        }
        confirmVerified(fileLogicMock)
    }

    @Test
    fun deleteById_simdutFileNotExists_doesNotCallDeleteInFileLogic() {
        // Arrange
        every { materialLogic.getById(any()) } returns material
        every { materialServiceMock.isUsedByMixMaterialOrMixType(any()) } returns false
        every { materialServiceMock.deleteById(any()) } just runs
        every { fileLogicMock.exists(any()) } returns false
        every { fileLogicMock.delete(any()) } just runs

        val simdutPath = Material.getSimdutFilePath(material.name)

        // Act
        materialLogic.deleteById(material.id)

        // Assert
        verify {
            fileLogicMock.exists(simdutPath)
        }
        verify(exactly = 0) {
            fileLogicMock.delete(simdutPath)
        }
        confirmVerified(fileLogicMock)
    }

    @Test
    fun deleteById_usedByMixMaterialOrMixType_throwsCannotDeleteException() {
        // Arrange
        every { materialLogic.getById(any()) } returns material
        every { materialServiceMock.isUsedByMixMaterialOrMixType(any()) } returns true
        every { materialServiceMock.deleteById(any()) } just runs
        every { fileLogicMock.exists(any()) } returns false
        every { fileLogicMock.delete(any()) } just runs

        // Act
        // Assert
        assertThrows<CannotDeleteException> { materialLogic.deleteById(material.id) }
    }
}